"""Declares primitives to parse data from the operating
system environment.
"""
from unimatrix.exceptions import ImproperlyConfigured


def parseint(environ, key, default):
    """Parse the environment variable as an integer."""
    try:
        return int(environ.get(key) or default)
    except ValueError:
        raise ImproperlyConfigured('%s can not be cast to integer.' % key)

def parsebool(environ, key):
    """Parses the environment variable into a boolean."""
    return str.lower(environ.get(key) or '') in ('1', 'yes', 'y', 'enabled')


def parselist(environ, key, sep=':', cls=tuple):
    """Parses the environment variable into an iterable,
    using `sep` as a separator e.g.::

    >>> from unimatrix.lib.environ import parselist
    >>> parselist({'foo': '1:2'}, 'foo')
    ('1', '2')
    """
    values = environ.get(key) or ''
    return cls(filter(bool, str.split(values, sep)))



# pylint: skip-file
import time
import unittest

from .. import timezone


class CurrentTimeTestCase(unittest.TestCase):

    def test_now_increases_over_time(self):
        t1 = timezone.now()
        time.sleep(0.1)
        t2 = timezone.now()
        self.assertGreater(t2, t1)



"""The :mod:`unimatrix.runtime` package provides a common interface to set
up the application runtime environment.
"""
import asyncio
import importlib
import inspect
import pkg_resources
import os
import warnings

from unimatrix.lib import meta
from . import logging


IS_CONFIGURED = False


def get_entrypoints(name):
    entrypoints = [
        (entry_point.name, entry_point.load())
        for entry_point
        in pkg_resources.iter_entry_points(name)
    ]
    return sorted(entrypoints, key=lambda x: getattr(x[1], 'WEIGHT', 0))


def get_settings_module(module_name: str) -> str:
    """Return a string containing the qualified Python name of the Unimatrix
    settings module. The qualified name is retrieved from the environment
    variable ``UNIMATRIX_SETTINGS_MODULE`` or from the standard location of
    ``<app symbol>.runtime.settings``.
    """
    return os.getenv('UNIMATRIX_SETTINGS_MODULE')\
        or f"{str.split(module_name, '.')[0]}.runtime.settings"


def execute(func, *args, **kwargs):
    """Setup the Unimatrix environment and run `func`."""
    async def f():
        await on('boot')
        try:
            return await func(*args, **kwargs)
        finally:
            await on('shutdown')
    return asyncio.run(f())


async def run_entrypoints(
    name: str,
    fn: str = None,
    namespace: str = 'unimatrix.runtime',
    *args, **kwargs) -> None:
    """Runs the entrpoints specified by `name`."""
    qualname = f'{namespace}.{name}'
    for _, entry_point in get_entrypoints(qualname):
        f = entry_point if not fn else getattr(entry_point, fn, None)
        if not callable(f):
            warnings.warn(
                f"Entrypoint {entry_point} (fn: {fn}) was not callable."
            )
            continue
        await f(*args, **kwargs)\
            if inspect.iscoroutinefunction(f)\
            else f(*args, **kwargs)


@meta.allow_sync
async def setup():
    """Setup all :mod:`unimatrix` components."""
    global IS_CONFIGURED
    if IS_CONFIGURED:
        return

    logging.configure()
    await run_entrypoints('init')
    for _, module in get_entrypoints('unimatrix.runtime'):
        on_setup = getattr(module, 'on_setup', None)
        if on_setup is None:
            continue
        await on_setup()\
            if inspect.iscoroutinefunction(on_setup)\
            else on_setup()

    IS_CONFIGURED = True


@meta.allow_sync
async def teardown():
    """Teardown all :mod:`unimatrix` components."""
    for _, module in get_entrypoints('unimatrix.runtime'):
        on_teardown = getattr(module, 'on_teardown', None)
        if on_teardown is None:
            continue
        await on_teardown()\
            if inspect.iscoroutinefunction(on_teardown)\
            else on_teardown()

    IS_CONFIGURED = False # pylint: disable=unused-variable
    await run_entrypoints('cleanup')


async def startup(*args, **kwargs):
    """Shorthand for ``unimatrix.runtime.on('boot')``."""
    return await on('boot', *args, **kwargs)


async def shutdown(*args, **kwargs):
    """Shorthand for ``unimatrix.runtime.on('shutdown')``."""
    return await on('teardown', *args, **kwargs)


@meta.allow_sync
async def on(name, *args, runlevel=None, **kwargs):
    if name == 'boot':
        await setup()
    elif name == 'shutdown':
        await teardown()

    # Before running the event handlers in UNIMATRIX_BOOT_MODULE, run the
    # entrypoints of installed modules.
    for _, module in get_entrypoints(f'unimatrix.runtime.{name}'):
        handler = getattr(module, name, None)
        if handler is None:
            continue
        await handler()\
            if inspect.iscoroutinefunction(handler)\
            else handler()

    qualname = os.getenv('UNIMATRIX_BOOT_MODULE')
    if qualname is None:
        return
    try:
        obj = importlib.import_module(qualname)
    except ImportError:
        warnings.warn(
            f"UNIMATRIX_BOOT_MODULE points to a non-existant symbol: "
            f"{qualname}.")
        return

    if not callable(obj) and not inspect.ismodule(obj):
        raise TypeError(
            "UNIMATRIX_BOOT_MODULE must point to a callable or module."
        )
    if callable(obj):
        return await obj(name, *args, **kwargs)\
            if inspect.iscoroutinefunction(obj)\
            else obj(name, *args, **kwargs)

    f = getattr(obj, name, None)
    if f is None:
        return
    return await f(*args, **kwargs)\
        if inspect.iscoroutinefunction(f)\
        else f(*args, **kwargs)



# pylint: skip-file
import unittest.mock


def sync_func(name, *args, **kwargs):
    return args, kwargs


async def func(name, *args, **kwargs):
    return args, kwargs


async def boot(*args, **kwargs):
    return args, kwargs


async def shutdown(*args, **kwargs):
    return args, kwargs



# pylint: skip-file
import unittest

from ..config import parse_environment


class ParseEnvironmentTestCase(unittest.TestCase):

    def test_unsupported_engine_raises_valueerror(self):
        env = {
            'DB_ENGINE': 'foo',
        }
        with self.assertRaises(ValueError):
            opts = parse_environment(env)

    def test_parse_autocommit_enabled(self):
        opts = parse_environment({})
        self.assertEqual(opts, None)
        env = {
            'DB_ENGINE': 'postgresql',
            'DB_AUTOCOMMIT': '1'
        }
        opts = parse_environment(env)
        self.assertTrue(opts.autocommit)

    def test_parse_autocommit_disabled(self):
        opts = parse_environment({})
        self.assertEqual(opts, None)
        env = {
            'DB_ENGINE': 'postgresql',
            'DB_AUTOCOMMIT': '0'
        }
        opts = parse_environment(env)
        self.assertFalse(opts.autocommit)

    def test_parse_postgresql_with_defaults(self):
        env = {
            'DB_ENGINE': 'postgresql',
        }
        opts = parse_environment(env)
        self.assertEqual(opts.host, 'localhost')
        self.assertEqual(opts.port, 5432)
        self.assertEqual(opts.name, None)
        self.assertEqual(opts.username, None)
        self.assertEqual(opts.password, None)

    def test_parse_mysql_with_defaults(self):
        env = {
            'DB_ENGINE': 'mysql',
        }
        opts = parse_environment(env)
        self.assertEqual(opts.host, 'localhost')
        self.assertEqual(opts.port, 3306)
        self.assertEqual(opts.name, None)
        self.assertEqual(opts.username, None)
        self.assertEqual(opts.password, None)

    def test_parse_sqlite_with_defaults(self):
        env = {
            'DB_ENGINE': 'sqlite',
        }
        opts = parse_environment(env)
        self.assertEqual(opts.host, None)
        self.assertEqual(opts.port, None)
        self.assertEqual(opts.name, ":memory:")
        self.assertEqual(opts.username, None)
        self.assertEqual(opts.password, None)

    def test_parse_mssql_with_defaults(self):
        env = {
            'DB_ENGINE': 'mssql',
        }
        opts = parse_environment(env)
        self.assertEqual(opts.host, 'localhost')
        self.assertEqual(opts.port, 1433)
        self.assertEqual(opts.name, None)
        self.assertEqual(opts.username, None)
        self.assertEqual(opts.password, None)
        self.assertEqual(opts.options.driver, "ODBC Driver 17 for SQL Server")
        self.assertEqual(opts.options.unicode_results, True)

    def test_parse_mssql_with_db_driver(self):
        env = {
            'DB_ENGINE': 'mssql',
            'DB_DRIVER': "ODBC Driver 13 for SQL Server"
        }
        opts = parse_environment(env)
        self.assertEqual(opts.host, 'localhost')
        self.assertEqual(opts.port, 1433)
        self.assertEqual(opts.name, None)
        self.assertEqual(opts.username, None)
        self.assertEqual(opts.password, None)
        self.assertEqual(opts.options.driver, "ODBC Driver 13 for SQL Server")
        self.assertEqual(opts.options.unicode_results, True)



# pylint: skip-file
import unittest

from ..dto import DTO
from ..dto import ImmutableDTO


class DataTransferObjectTestCase(unittest.TestCase):
    dto_class = DTO

    def test_from_dict_recursive(self):
        mapping = {
            'foo': {'bar': 1}
        }
        dto = self.dto_class.fromdict(mapping)
        self.assertIsInstance(dto, self.dto_class)
        self.assertIsInstance(dto['foo'], self.dto_class)

    def test_serialize_returns_dict(self):
        d1 = self.dto_class(foo=1)
        self.assertIsInstance(d1.__serialize__(lambda x: x), dict)

    def test_as_dict_nested_dto(self):
        d1 = self.dto_class(
            foo=self.dto_class(bar=1)
        )
        d2 = d1.as_dict()
        self.assertIsInstance(d2['foo'], dict)

    def test_as_dict_nested_dict(self):
        d1 = self.dto_class(
            foo=dict(bar=1)
        )
        d2 = d1.as_dto()
        self.assertIsInstance(d2['foo'], self.dto_class)

    def test_as_dto_with_dto_recurse(self):
        d1 = self.dto_class(
            foo=dict(
                bar=self.dto_class(baz=1)
            )
        )
        d2 = d1.as_dto(with_dto=True)
        self.assertIsInstance(d2.foo.bar, self.dto_class)

    def test_list_of_dicts_to_dtos(self):
        d1 = self.dto_class(
            foo=[
                {'bar': 1},
                {'baz': 2}
            ])
        d2 = d1.as_dto()
        self.assertIsInstance(d2.foo[0], self.dto_class)
        self.assertIsInstance(d2.foo[1], self.dto_class)

    def test_setattr(self):
        if self.dto_class.__immutable__:
            raise unittest.SkipTest("No setter test on immutable.")
        d1 = self.dto_class(foo=1)
        d1.foo = 2
        self.assertEqual(d1.foo, 2)

    def test_getattr_raises_on_nonexisting(self):
        d1 = self.dto_class(foo=1)
        with self.assertRaises(AttributeError):
            d1.bar

    def test_intersect_union(self):
        d1 = self.dto_class(foo=1)
        d2 = self.dto_class(foo=1, bar=2)
        self.assertEqual(set((d1 & d2).keys()), set(['foo']))

    def test_intersect_with_set(self):
        d1 = self.dto_class(foo=1)
        d2 = d1 & ['foo']
        self.assertEqual(set((d1 & d2).keys()), set(['foo']))

    def test_intersect_takes_right_operand_values(self):
        d1 = self.dto_class(foo=1)
        d2 = self.dto_class(foo=2, bar=2)
        d3 = d1 & d2
        self.assertEqual(d3.foo, 2)

    def test_union_non_dict_raises_valuerror(self):
        d1 = self.dto_class(foo=1)
        with self.assertRaises(ValueError):
            d1 | ['foo']

    def test_union_takes_values_from_right_operand(self):
        d1 = self.dto_class(foo=1)
        d2 = self.dto_class(foo=2, bar=3)
        d3 = d1 | d2
        self.assertEqual(d2.foo, 2)

    def test_union_takes_all_keys(self):
        d1 = self.dto_class(foo=1)
        d2 = self.dto_class(foo=2, bar=3)
        d3 = d1 | d2
        self.assertEqual(set(d3.keys()), set(['foo','bar']))

    def test_sub(self):
        d1 = self.dto_class(foo=1, bar=2)
        d2 = d1 - ['bar']
        self.assertEqual(set(d2.keys()), set(['foo']))

    def test_sub_raises_with_non_list(self):
        d1 = self.dto_class(foo=1, bar=2)
        with self.assertRaises(TypeError):
            d1 - 'foo'

    def test_string_representiation(self):
        self.assertEqual(repr(self.dto_class(foo=1)),
            "<%s: {'foo': 1}>" % self.dto_class.__name__)


class ImmutableDataTransferObjectTestCase(DataTransferObjectTestCase):
    dto_class = ImmutableDTO

    def test_setattr(self):
        if not self.dto_class.__immutable__:
            raise unittest.SkipTest("No setter test on non-immutable.")
        d1 = ImmutableDTO(foo=1)
        with self.assertRaises(TypeError):
            d1.foo = 2



"""Declares :class:`Pattern`."""
import itertools


class Pattern:
    """Represents a matching pattern, possible containing wildcards."""

    @staticmethod
    def compare(pattern, value, sep='.'):
        """Compare `pattern` against `value` using the separator character
        `sep`.
        """
        if pattern == value:
            return True

        eq = None
        for x, y in itertools.zip_longest(
            str.split(pattern, sep),
            str.split(value, sep)
        ):
            if None in (x, y):
                # If any of the items is None, it means that the patterns
                # did not have the same length.
                eq = False
                break
            if (x == y) or (x == '*'):
                continue
            eq = False
            break
        else:
            eq = True
        assert eq is not None # nosec
        return eq

    def __init__(self, pattern: str, sep: str):
        self.pattern = pattern
        self.sep = sep

    def is_match(self, value: str) -> bool:
        """Return a boolean indicating if the pattern matches the value."""
        result = self.compare(self.pattern, value, self.sep)
        print(self.pattern, value, result)
        return result

    def __str__(self) -> str:
        return self.pattern

    def __eq__(self, value) -> bool:
        return self.is_match(value)

    def __hash__(self) -> int:
        return hash(self.pattern)



"""Provides testing helper functions and classes."""
import asyncio
import functools
import inspect
import os
import types
import unittest

from unimatrix.lib.environ import parselist


__all__ = ['needs']


#: The current testing stage as defined by the operating system environment
#: variable ``TEST_STAGE``.
TEST_STAGE = os.getenv('TEST_STAGE')

#: The current testing scope as defined by the operating system environment
#: variable ``TEST_SCOPE``.
TEST_SCOPE = os.getenv('TEST_SCOPE')

#: Used to specify `unit` tests.
TEST_UNIT = 'unit'

#: Used to specify `integration` tests.
TEST_INTEGRATION = 'integration'

#: Used to specify `system` tests.
TEST_SYSTEM = 'system'


def needs(capabilities):
    """Return a decorator that skips the test if the required capabilities
    are not satisfied by the test environment. For a description of
    allowed values, refer to :envvar:`TEST_STACK_CAPABILITIES`.

    Args:
        capabilities (list): the list of capabilities required by
            the decorated test.

    A test can declare its required capabilities with the :func:`needs`
    decorator. The decorator may be used at the class or at the
    function level:

    .. code:: python

        class DatabaseTestCase(unittest.TestCase):

            @unimatrix.lib.test.needs('rdbms:postgres')
            def test_run_on_any_version(self):
                ... tests

            @unimatrix.lib.test.needs('rdbms:postgres:13')
            def test_run_on_postgres_13(self):
                ... tests
    """
    if isinstance(capabilities, str):
        capabilities = [capabilities]
    required = set(capabilities)
    provided = parselist(os.environ, 'TEST_STACK_CAPABILITIES', sep=';', cls=set)

    # Quite inefficient
    for cap in set(provided):
        for i in range(len(str.split(cap, ':'))):
            parent = str.join(':', str.rsplit(cap, ':', i+1)[:-i+1])
            if not parent:
                break
            provided.add(parent)

    missing = list(sorted(required - provided))
    if not missing:
        return unittest.skipIf(False, "Required capabilities satisfied.")

    return unittest.skipIf(bool(missing),
        f"Capabilities not satisfied: {','.join(missing)}")


class TestCaseSelector:
    """A class, method or function decorator that skips tests based on the
    configured testing environment.

    Args:
        name (:obj:`str`): the :term:`Test Stage` in which the decorated
            callable must run.
        scopes (:obj:`list`): a list of scopes in which the decorated callable
            must run. If `scopes` is ``None`` and a scope is defined, then the
            test is skipped.
    """

    def __init__(self, name, scopes=None, current_stage=None, current_scope=None):
        self._stage = name
        self._scopes = scopes
        self._current_stage = current_stage
        self._current_scope = current_scope

    def get_current_scope(self):
        """Returns the current testing scope or ``None``."""
        return self._current_scope or TEST_SCOPE

    def get_current_stage(self):
        """Returns the current testing stage or ``None``."""
        return self._current_stage or TEST_STAGE

    def scopes(self, scopes):
        """Add `scopes` to the testing scopes in which the decorated test
        must run.
        """
        if scopes is None:
            return self
        if self._scopes is None:
            self._scopes = []
        self._scopes.extend(scopes)
        return self

    def scope(self, scope):
        """Like :meth:`scopes`, but with a single positional argument of
        type :obj:`str`. Use when only a single scope is to be indicated.
        """
        return self.scopes([scope])

    def skip(self, message): #pylint: disable=no-self-use
        """Ensure that the decorated object is skipped."""
        raise unittest.SkipTest(message)

    def must_skip(self):
        """Return a :obj:`tuple` containing a :obj:`bool` and :obj:`str`
        indicating if a test must be skipped under the current environment,
        and the reason.
        """
        reason_stage = "Test only runs in the '%s' stage" % self._stage
        reason_scope = "Test only runs in the '%s' scope" % self._stage
        current_stage = self.get_current_stage()
        current_scope = self.get_current_scope()
        must_skip = False
        reason = None
        if (current_stage != self._stage)\
        or not current_stage:
            must_skip = True
            reason = reason_stage

        if (current_scope and current_scope not in (self._scopes or []))\
        or (not current_scope and self._scopes):
            must_skip = True
            reason = reason_scope

        return must_skip, reason


    def __call__(self, decoratable):
        """Decorates a decoratable object (class, method or function) and
        raises :exc:`unittest.SkipTest` if the decorated object must be
        skipped by the test runner in the current environment.
        """
        return self._decorate_type(decoratable)\
            if not isinstance(decoratable, types.FunctionType)\
            else self._decorate_function(decoratable)

    def _decorate_type(self, cls):
        must_skip, reason = self.must_skip()
        if must_skip:
            cls.__unittest_skip__ = True
            cls.__unittest_skip_why__ = reason
        return cls

    def _decorate_function(self, func):
        must_skip, reason = self.must_skip()
        if must_skip:
            @functools.wraps(func)
            def decorator(*args, **kwargs): #pylint: disable=unused-argument
                self.skip(reason)
            func = decorator
        return func


def async_test(func):
    """Decorator to run a test asynchronously."""
    @functools.wraps(func)
    def f(self, *args, **kwargs):
        return self.loop.run_until_complete(
            func(self, *args, **kwargs))

    return f


class AsyncTestCaseMetaclass(type):

    def __new__(cls, name, bases, attrs):
        super_new = super().__new__
        if name == 'AsyncTestCase':
            return super_new(cls, name, bases, attrs)

        # Ensure that the loop is created on each run of the test class.
        setup = attrs.pop('setUp', None) or (lambda self: None)
        def setUp(self):
            self.loop = asyncio.new_event_loop()
            if inspect.iscoroutinefunction(setup):
                f = lambda self: self.loop.run_until_complete(setup(self))
            else:
                f = setup
            return f(self)
        attrs['setUp'] = setUp

        for attname, function in list(dict.items(attrs)):
            if not inspect.iscoroutinefunction(function)\
            or not (str.startswith(attname, 'test_') or attname == 'tearDown'):
                continue
            attrs[attname] = async_test(function)

        return super_new(cls, name, bases, attrs)


class AsyncTestCase(unittest.TestCase, metaclass=AsyncTestCaseMetaclass):
    pass


unit = TestCaseSelector(TEST_UNIT) #pylint: disable=invalid-name
integration = TestCaseSelector(TEST_INTEGRATION) #pylint: disable=invalid-name
system = TestCaseSelector(TEST_SYSTEM) #pylint: disable=invalid-name
